#ifndef _HASH_TABLE_H
#define _HASH_TABLE_H

#include "globals.h"
#include "util.h"

using namespace std;

class HashTable {
  
 private:
  ValueType     *valueTable;
  uint64         a0, b0, a1, b1; // used for 2-universal hashing
  uint32         numBins;        // 1<<inxBits
  uint32         binMask;        // numBins-1

 public:
  // Constructor
  HashTable(int seed, int inxBits);

  // Destructor
  ~HashTable();

  // Accessors
  uint32 GetNumBins() const { return numBins; }

  // Member Functions
  void Reset()
    {
      for ( uint32 i = 0; i < numBins; i++ )
        valueTable[i] = 0;
    }
  
  void AddValue(KeyType key, ValueType val)
    {
      valueTable[GetIndex(key)] += val;
    }
  
  void RemoveValue(IndexType key, ValueType val)
    {
      valueTable[GetIndex(key)] -= val;
    }
  
  ValueType GetValue(KeyType key) 
    {
      return valueTable[GetIndex(key)];
    }

  // 2-universal hashing via Dietzfelbinger
__inline__  IndexType GetIndex(uint32 key)
    {
      return smartdietz(key, binMask, a0, b0);
    }

  // 2-universal hashing via Dietzfelbinger
__inline__  IndexType GetIndex(uint64 key)
    {
      return smartdietz(key, binMask, a0, b0, a1, b1);
    }
};

#endif
