#ifndef _REV_K_SKETCH_48_H
#define _REV_K_SKETCH_48_H

#include <math.h>
#include <iostream>
#include <vector>
#include <list>
#include <algorithm>
#include "revHash48.h"
#include "kSketch.h"
#include "mangler.h"
#include "globals.h"
//#include "revKSketch.h"

using namespace std;

    
typedef struct NodeType {
  list<unsigned char> keyList;
  unsigned char numFound;  // Number of HT's in which this key has been found

  // List of heavy change buckets to which this key corresponds for each HT
  vector<list<IndexType> > bktLists;

  NodeType(unsigned char key, IndexType bkt, 
	   unsigned char tbl, int numTbls)
  {
    keyList.push_back(key);
    numFound = 1;
    
    bktLists.resize(numTbls);
    bktLists[tbl].push_back(bkt);
  }

  NodeType()
  {
    numFound = 0;
  }

} NodeType;

  
typedef vector<list<NodeType*> > GraphType;

class RevKSketch48 {

  // private:
 public:
  vector<RevHashTable48> hashTbls;
  KSketch *extraSketch;
  
  int numTbls;
  int r;
  int iterSize;  // Number of changes to be included in 1 iteration

  ValueType sum;

  // IP Mangling additions
#ifdef NEW_MANGLE
  Mangler *mangler;
#else
  unsigned int a1, a2;
  unsigned int a1_inv, a2_inv;
  unsigned int FindInverse(unsigned int a);
#endif

  KeyType48 ReorderKey(KeyType48 key);
  KeyType48 RestoreKey(KeyType48 key);

  // Helper Functions for Change Detection

  // CreateNodes creates the nodes of the graph. The graph consists
  // of a 2D matrix where the rows are the key segments and the columns
  // are the individual nodes for that particular segment.  The input
  // into this function is a list of the heavy change buckets for each 
  // of the hash tables.
  void CreateNodes(GraphType &graph, 
		   const vector<list<IndexType> > &bktLists);

  // CreateLinks finds all of the valid paths between two levels in the
  // graph.  The nodes in the second level of the graph are replaced
  // by super nodes representing the valid paths.  The contents of both
  // original lists are destroyed.
  void CreateLinks(GraphType &graph, int startDiv);

  // RIntersect takes two nodes and intersects the 2D matrices.  If
  // more than R rows have a NULL intersection, than the R-Intersection
  // is stopped and NULL is returned.  If the last row is intersected,
  // a new node containing this intersection is created.  The key
  // becomes a concatenated list of node1's key followed by node2's key.
  NodeType* RIntersect(NodeType *node1, NodeType *node2);

  void GenerateFullKeys(GraphType &graph, list<KeyType48> &resultList);

 public:
  RevKSketch48(int numHashTbls, int myR, int inxBits, int mySize);
  ~RevKSketch48();

  void AddValue(KeyType48 key, ValueType val);
  void RemoveValue(KeyType48 key, ValueType val);
  ValueType EstValue(KeyType48 key);

  // Debug Function
  IndexType GetIndex(KeyType48 key, int tbl);

   void ListIntersect(list<IndexType> &outList,
		   const list<IndexType> &l1,
		   const list<IndexType> &l2);
  // This function will populate the result list with the IP
  // addresses that are associated with changes larger than the
  // given threshold.
  //
  // Changes are computed by: thisSketch - otherSketch
  // No absolute value is taken.
  void ChangeDetection(list<KeyType48> &resultList,
		       RevKSketch48 &otherSketch, double threshold);
};

#endif
